#!/usr/local/bin/python3.9

import os 
import time
dir_path = os.path.dirname(os.path.realpath(__file__))
os.chdir(dir_path)
print(dir_path)

# minikube kubectl -- create -f ../start.yaml ; sleep 60 ; minikube service pf8087-csu-csip-oms -n csip

import requests

import tkinter as tk
import tkinter.messagebox
import customtkinter
import json
import os
from PIL import Image, ImageTk
import traceback
from multiprocessing import Process
from multiprocessing import Queue
import re
import pandas as pd
import numpy as np
import ast
import platform
import time

from ..util import PSORunner
from ..util import GraphGenerator

import subprocess
import plotly.express as px
import plotly.graph_objs as go

from tkinter.filedialog import askopenfilename
from tkinter.filedialog import asksaveasfilename

from . import OptionManager as om

from .SetupTab import SetupTab as st
from .PlatformTab import PlatformTab as pt
from .RunTab import RunTab as rt
from .VisualizeTab import VisualizeTab as vt

from ..util.CTkToolTip import CTkToolTip as ctt

customtkinter.set_appearance_mode("Dark")  # Modes: "System" (standard), "Dark", "Light"
customtkinter.set_default_color_theme("blue")  # Themes: "blue" (standard), "green", "dark-blue"

class App(customtkinter.CTk):
    def __init__(self):
        super().__init__()
        
        self.option_manager = om.OptionManager()
        
        self.running_config = None
        self.selected_graph_name = None

        self.train_process = None
        self.minikube_process = None
        self.data_x = [0]
        self.data_y = [0]
        
        self.image_width = 1280
        self.image_height = 720

        # configure window
        self.title("CSIP PSO")
        self.geometry(f"{1920}x{1080}")

        # configure grid layout (4x4)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(1, weight=1)

        header_padding_x = (5, 5)
        header_padding_y = (10, 10)
        
        self.sidebar_frame = customtkinter.CTkFrame(self, width=140, corner_radius=0)
        self.sidebar_frame.grid(row=0, column=0, sticky="nsew")
        self.sidebar_frame.grid_columnconfigure(4, weight=1)
        self.logo_label = customtkinter.CTkLabel(self.sidebar_frame, text="CSIP PSO", font=customtkinter.CTkFont(size=20, weight="bold"))
        self.logo_label.grid(row=0, column=0, padx=(20, 10), pady=header_padding_y)
        self.save_button = customtkinter.CTkButton(self.sidebar_frame, text="Save", width=60, command=self.save_project)
        self.save_button.grid(row=0, column=1, padx=header_padding_x, pady=header_padding_y)
        self.load_button = customtkinter.CTkButton(self.sidebar_frame, text="Load", width=60, command=self.load_project)
        self.load_button.grid(row=0, column=2, padx=header_padding_x, pady=header_padding_y)
        
        # 4 is service URL
        self.service_label = customtkinter.CTkLabel(self.sidebar_frame, text="Service:", anchor="w")
        self.service_label.grid(row=0, column=3, padx=(80, 5), pady=header_padding_y)
        self.service_url = customtkinter.CTkEntry(self.sidebar_frame, textvariable=self.option_manager.get_arguments()['url'])
        self.service_url.grid(row=0, column=4, columnspan=1, padx=header_padding_x, pady=header_padding_y, sticky="nsew")
        refresh_image = customtkinter.CTkImage(Image.open(os.path.join("./images", "refresh.png")), size=(20, 20))
        self.refresh_button = customtkinter.CTkButton(self.sidebar_frame, text=None, width=30, image=refresh_image, command=self.load)
        self.refresh_button.grid(row=0, column=5, padx=(5, 80), pady=header_padding_y)
        
        self.scaling_label = customtkinter.CTkLabel(self.sidebar_frame, text="Scale:", anchor="w")
        self.scaling_label.grid(row=0, column=6, padx=header_padding_x, pady=header_padding_y)
        self.scaling_optionemenu = customtkinter.CTkOptionMenu(self.sidebar_frame, values=["50%", "75%", "100%", "125%", "150%", "175%", "200%"], width=60,
                                                               command=self.change_scaling_event)
        self.scaling_optionemenu.grid(row=0, column=7, padx=header_padding_x, pady=header_padding_y)
        self.scaling_optionemenu.set("100%")
        
        expand_image = customtkinter.CTkImage(Image.open(os.path.join("./images", "expand.png")), size=(20, 20))
        self.new_window = customtkinter.CTkButton(self.sidebar_frame, text=None, width=30, image=expand_image, command=self.new_window)
        self.new_window.grid(row=0, column=8, padx=(5, 20), pady=header_padding_y)
        
        self.tabview = customtkinter.CTkTabview(self, bg_color="transparent", fg_color="transparent")
        self.tabview.grid(row=1, column=0, padx=(0, 0), pady=(10, 10), sticky="nsew")
        tab1 = "Platform"
        tab2 = "Setup"
        tab3 = "Run"
        tab4 = "Visualize"
        tab5 = "Results"
        
        self.tabview.add(tab1)
        self.tabview.add(tab2)
        self.tabview.add(tab3)
        self.tabview.add(tab4)
        self.tabview.add(tab5)
        
        #self.tabview.configure(state="disabled")
        
        pt.create_tab(self, self.tabview.tab(tab1))
        st.create_tab(self, self.tabview.tab(tab2))
        rt.create_tab(self, self.tabview.tab(tab3))
        vt.create_tab(self, self.tabview.tab(tab4))
        
        self.footer_frame = customtkinter.CTkFrame(self, corner_radius=0)
        self.footer_frame.grid(row=2, column=0, sticky="nsew")
        self.footer_frame.grid_columnconfigure(4, weight=1)
        
        self.footer_progress_label = customtkinter.CTkLabel(self.footer_frame, text="Stopped", width=150, font=customtkinter.CTkFont(size=16, weight="bold"), anchor="w")
        self.footer_progress_label.grid(row=0, column=0, padx=(20, 5), pady=header_padding_y)

        self.footer_progress_bar = customtkinter.CTkProgressBar(self.footer_frame)
        self.footer_progress_bar.grid(row=0, column=4, padx=(50, 100), pady=header_padding_y, sticky="ew")
        self.footer_progress_bar.set(0)
        
        play_image = customtkinter.CTkImage(Image.open(os.path.join("./images", "play.png")), size=(20, 20))
        self.new_window = customtkinter.CTkButton(self.footer_frame, text=None, width=30, image=play_image, command=self.run)
        self.new_window.grid(row=0, column=7, padx=(5, 5), pady=header_padding_y)
        
        stop_image = customtkinter.CTkImage(Image.open(os.path.join("./images", "stop.png")), size=(20, 20))
        self.new_window = customtkinter.CTkButton(self.footer_frame, text=None, width=30, image=stop_image, command=self.stop)
        self.new_window.grid(row=0, column=8, padx=(5, 20), pady=header_padding_y)

    def _resize_image(self, event):
        self.graph_label.update_idletasks()
        new_width = self.graph_label.winfo_width()
        new_height = self.graph_label.winfo_height()
        
        alt_width = new_height * 1.77778
        alt_height = new_width / 1.77778
        
        if (new_width < new_height):
            new_height = alt_height
        else:
            new_width = alt_width
        
        self.image_width = new_width
        self.image_height = new_height
        
        self.graph_image = customtkinter.CTkImage(self.graph_image_obj, size=(new_width, new_height))
        self.graph_label.configure(image=self.graph_image)
        self.graph_label.update_idletasks()

    def callback_test(self, *args):
        print("callback_test called")
            
    def update_graph(self, value):
        selected_graph = self.graph_selector_value.get()
        info = self.option_manager.get_project_data()
        folder = os.path.join(info['path'], info['name'])
        
        if not os.path.exists(folder):
            os.makedirs(folder)
        
        if (selected_graph == "Best Cost Stacked"):
            self.selected_graph_name = "best_cost_stacked"
            image_path = os.path.join(folder, self.selected_graph_name + ".png")
            if not os.path.exists(image_path):
                image_path = os.path.join("./images", "up.png")
            self.graph_image_obj = Image.open(image_path)
            self.graph_image = customtkinter.CTkImage(self.graph_image_obj, size=(self.image_width, self.image_height))
            self.graph_label.configure(image=self.graph_image)
        elif (selected_graph == "Best Cost by Round"):
            self.selected_graph_name = "best_cost_by_round"
            image_path = os.path.join(folder, self.selected_graph_name + ".png")
            if not os.path.exists(image_path):
                image_path = os.path.join("./images", "up.png")
            self.graph_image_obj = Image.open(image_path)
            self.graph_image = customtkinter.CTkImage(self.graph_image_obj, size=(self.image_width, self.image_height))
            self.graph_label.configure(image=self.graph_image)
        elif (selected_graph == "Table"):
            self.selected_graph_name = "table"
            image_path = os.path.join(folder, self.selected_graph_name + ".png")
            if not os.path.exists(image_path):
                image_path = os.path.join("./images", "up.png")
            self.graph_image_obj = Image.open(image_path)
            self.graph_image = customtkinter.CTkImage(self.graph_image_obj, size=(self.image_width, self.image_height))
            self.graph_label.configure(image=self.graph_image)
        elif (selected_graph == "Calibrated Parameters"):
            self.selected_graph_name = "calibrated_params_by_round"
            image_path = os.path.join(folder, self.selected_graph_name + ".png")
            if not os.path.exists(image_path):
                image_path = os.path.join("./images", "up.png")
            self.graph_image_obj = Image.open(image_path)
            self.graph_image = customtkinter.CTkImage(self.graph_image_obj, size=(self.image_width, self.image_height))
            self.graph_label.configure(image=self.graph_image)
            
    def save_project(self):
        metrics = self.option_manager.get_metrics()
        filename = asksaveasfilename(filetypes=[("JSON", "*.json")], initialfile="config", defaultextension="json", title="Save Project")
        
        try:
        
            # Convert metrics to json and save to file with proper spacing
            with open(filename, "w") as f:
                f.write(json.dumps(metrics, indent=4))
                
            self.save_button.configure(text="Saved!")
            self.after(3000, lambda: self.save_button.configure(text="Save"))
        except Exception as e:
            self.save_button.configure(text="Error!")
            print(e)
            self.after(3000, lambda: self.save_button.configure(text="Save"))
    
    def new_window(self):
        # Shell out and run ./main.py
        subprocess.Popen(["python3", "../mgpsogui.py"])
    
    def load_project(self):
        
        filename = askopenfilename(filetypes=[("JSON", "*.json")], title="Open Project", multiple=False)
        print(filename)
        
        try:
        
            # Load config.json and convert to metrics
            with open(filename, "r") as f:
                metrics = json.loads(f.read())
            
            self.option_manager.set_path(filename)
            
            if "arguments" in metrics:
                metrics["arguments"]["calibration_parameters"] = metrics["calibration_parameters"]
            
            if "service_parameters" in metrics:
                self.option_manager.set_service_parameters(metrics["service_parameters"])
                self.tabview.configure(state="enabled")
            
            print(metrics)
            
            self.option_manager.clear()
            self.option_manager.add_arguments(metrics["arguments"])
            self.option_manager.add_steps(metrics["steps"])
            
            self.steps_frame.clear()
            self.steps_frame.render()

            self.static_param_frame.clear()
            self.static_param_frame.render()
            
            self.calib_param_frame.clear()
            self.calib_param_frame.render()
            
            self.load_button.configure(text="Loaded!")
            self.after(3000, lambda: self.load_button.configure(text="Load"))
            
        except Exception as e:
            print(e)
            self.load_button.configure(text="Error!")
            self.after(3000, lambda: self.load_button.configure(text="Load"))

    def change_appearance_mode_event(self, new_appearance_mode: str):
        customtkinter.set_appearance_mode(new_appearance_mode)

    def change_scaling_event(self, new_scaling: str):
        new_scaling_float = int(new_scaling.replace("%", "")) / 100
        customtkinter.set_widget_scaling(new_scaling_float)
        
    def change_scaling_event(self, new_scaling: str):
        new_scaling_float = int(new_scaling.replace("%", "")) / 100
        customtkinter.set_widget_scaling(new_scaling_float)
        
    def make_request(self):
        service_url = self.service_url.get()
        try:
            response = requests.get(service_url)
                
            response_json = json.loads(response.text)
            status = response.status_code
            
            self.option_manager.set_service_parameters(response_json)
            
            self.service_status.delete('0.0', tk.END)
            self.service_status.insert(text=str(status), index='0.0')
            self.service_name.delete('0.0', tk.END)
            self.service_name.insert(text=str(response_json["metainfo"]["name"]), index='0.0')
            self.service_description.delete('0.0', tk.END)
            self.service_description.insert(text=str(response_json["metainfo"]["description"]), index='0.0')
            self.service_details.delete('0.0', tk.END)
            self.service_details.insert(text=json.dumps(response_json, indent=4), index='0.0')
            
            self.refresh_button.configure(fg_color="green")
        except Exception as e:
            self.refresh_button.configure(fg_color="red")
        
    
    def load(self):
        # Make HTTP request to service_url and save the result to bounds.json
        
        self.default_button_color = self.refresh_button.cget("fg_color")
        
        self.refresh_button.configure(fg_color="gray")
        
        self.after(10, self.make_request)
        self.after(3000, lambda: self.refresh_button.configure(fg_color=self.default_button_color))
        
    def run(self):
        metrics = self.option_manager.get_metrics()
        self.running_config = metrics
        
        #self.progress_bar.configure(mode="indeterminnate")
        #self.progress_bar.start()
        self.footer_progress_bar.configure(mode="indeterminnate")
        self.footer_progress_bar.start()
        
        self.data_x = [0]
        self.data_y = [0]
        
        self.progress_message_middle.configure(text="Calibration starting...")
        self.footer_progress_label.configure(text="Starting...")
        
        self.textbox.insert("0.0", "Starting calibration...\n\n")
        self.textbox.insert("0.0", "Calibration Parameters:\n")
        self.textbox.insert("0.0", json.dumps(metrics, indent=4) + "\n\n")
        try:
            info = self.option_manager.get_project_data()
            folder = os.path.join(info['path'], info['name'])
            
            
            self.train_process = Process(target=PSORunner.run_process, args=(metrics, folder))
            self.train_process.daemon = True
            self.train_process.start()
            self.after(1000, self.watch_loop)
            self.string_cache = ""
            self.data_cache = ""
            
        except Exception as e:
            self.textbox.insert("0.0", "An exception occurred!\n Exception: " + str(e) + "\n\n")
            self.textbox.insert("0.0", "Stack trace:\n")
            self.textbox.insert("0.0", traceback.format_exc())
            self.textbox.insert("0.0", "\n\n")
            self.textbox.insert("0.0", "Calibration failed!")
            self.progress_message_left.configure(text="")
            self.progress_message_middle.configure(text="Calibration failed! See error log below.")
            self.progress_message_right.configure(text="")
            self.footer_progress_label.configure(text="Failed")
            #self.progress_bar.stop()
            self.footer_progress_bar.stop()
            #self.progress_bar.configure(mode="determinate")
            self.footer_progress_bar.configure(mode="determinate")
            #self.progress_bar.set(0)
            self.footer_progress_bar.set(0)
            
    def stop(self):
        print("Stopping...")
        self.train_process.terminate()
        
        info = self.option_manager.get_project_data()
        folder = os.path.join(info['path'], info['name'])
            
        if not os.path.exists(folder):
            os.makedirs(folder)
        
        # Stop the process
        if (os.path.exists(os.path.join(folder, 'output.txt'))):
            os.remove(os.path.join(folder, 'output.txt'))
        
        if (os.path.exists(os.path.join(folder, 'error.txt'))):
            os.remove(os.path.join(folder, 'error.txt'))
            
        self.textbox.insert("0.0", "\nCalibration terminated!\n")
        #self.progress_bar.stop()
        self.footer_progress_bar.stop()
        #self.progress_bar.configure(mode="determinate")
        self.footer_progress_bar.configure(mode="determinate")
        #self.progress_bar.set(0)
        self.footer_progress_bar.set(0)
        self.progress_message_left.configure(text="")
        self.progress_message_middle.configure(text="Calibration stopped!")
        self.footer_progress_label.configure(text="Stopped")
        self.progress_message_right.configure(text="")
        
    def watch_loop(self):
        # Check if file exists:
        
        info = self.option_manager.get_project_data()
        folder = os.path.join(info['path'], info['name'])
            
        if not os.path.exists(folder):
            os.makedirs(folder)
        
        if (os.path.exists(os.path.join(folder, 'output.txt'))):
            with open(os.path.join(folder, 'output.txt'), 'r') as f:
                lines = f.readlines()
                lines_string = "".join(lines)
                
                new_characters = lines_string.replace(self.string_cache, "")
                # Update the textbox with characters not in self.string_cache
                self.textbox.insert('0.0', new_characters)
                self.string_cache = lines_string
                print(new_characters, end="")
                
            try:
                with open(os.path.join(folder, "output.txt"), "r") as f:
                    text = f.read()

                calibrated_params_pattern = r"calibrated params: ({.*?})"
                best_particle_values_pattern = r"best particle values: (\[.*?\])"
                progress_pattern = r"Progress -  best_round_cost:(.*?), rel_round_tol:(.*?), rtol:(.*?)\n"

                calibrated_params = re.findall(calibrated_params_pattern, text)
                best_particle_values = re.findall(best_particle_values_pattern, text)
                progress_values = re.findall(progress_pattern, text)

                for index, pp in enumerate(best_particle_values):
                    pp = pp.strip()
                    pp = pp.replace('[ ', '[')
                    pp = pp.replace('  ', ',')
                    pp = pp.replace(' ', ',')
                    best_particle_values[index] = pp

                calibrated_params = [ast.literal_eval(i) for i in calibrated_params]
                best_particle_values = [ast.literal_eval(i) for i in best_particle_values]
                progress_values = [tuple(map(float, i)) for i in progress_values]
                
                print(calibrated_params)
                
                GraphGenerator.calibrated_params_by_round(self.running_config['steps'], calibrated_params, self.option_manager)
                self.update_graph("")
            except Exception as e:
                # Print stack trace
                traceback.print_exc()
                
                print(e)
                
        if (os.path.exists(os.path.join(folder, 'error.txt'))):
            with open(os.path.join(folder, 'error.txt'), 'r') as f:
                lines = f.readlines()
                lines_string = "".join(lines)
                self.data_cache = lines_string
                
                pattern = r'(\d+)%\|.*\|(\d+)/(\d+)(?:,\sbest_cost=(\d+\.\d+))?' # The magic of AI
                matches = re.findall(pattern, self.data_cache)
                filtered_matches = [match for match in matches if match[3] != '']
                matches = filtered_matches

                if len(matches) > 0:
                    df = pd.DataFrame(matches, columns=['percent', 'completed_rounds', 'total_rounds', 'best_cost'], dtype=float)
                    df = df[df['best_cost'] != '']
                    df = df.dropna()
                    df = df.drop_duplicates()
                    df['round_step'] = (df['completed_rounds'].diff() < 0).cumsum()
                    df = df.drop_duplicates(subset=['completed_rounds', 'round_step'])

                    GraphGenerator.best_cost_stacked(self.running_config['steps'], df, self.option_manager)
                    GraphGenerator.best_cost_by_round(self.running_config['steps'], df, self.option_manager)
                    GraphGenerator.table(self.running_config['steps'], df, self.option_manager)
                    self.update_graph("")

                    match = matches[-1]
                    percent = int(match[0])
                    completed_rounds = int(match[1])
                    total_rounds = int(match[2])
                    best_cost = float(match[3]) if match[3] else None
                    
                    if (percent > 0):
                        #self.progress_bar.stop()
                        self.footer_progress_bar.stop()
                        #self.progress_bar.configure(mode="determinate")
                        self.footer_progress_bar.configure(mode="determinate")
                        #self.progress_bar.set(percent/100)
                        self.footer_progress_bar.set(percent/100)
                        self.progress_message_left.configure(text="Percent Complete: " + str(percent) + "%")
                        self.progress_message_middle.configure(text=str(completed_rounds) + "/" + str(total_rounds))
                        self.footer_progress_label.configure(text=str(completed_rounds) + "/" + str(total_rounds))
                        self.progress_message_right.configure(text="Best Cost: " + str(best_cost))
                    else:
                        #self.progress_bar.configure(mode="indeterminate")
                        self.footer_progress_bar.configure(mode="indeterminate")
                        #self.progress_bar.start()
                        self.footer_progress_bar.start()
                        self.progress_message_left.configure(text="")
                        self.progress_message_middle.configure(text="Starting new round...")
                        self.footer_progress_label.configure(text="Continuing...")
                        self.progress_message_right.configure(text="")
                
                
                
        if self.train_process.is_alive():
            self.after(1000, self.watch_loop)
        else:
            #self.progress_bar.stop()
            self.footer_progress_bar.stop()
            #self.progress_bar.configure(mode="indeterminate")
            self.footer_progress_bar.configure(mode="indeterminate")
            #self.progress_bar.start()
            self.footer_progress_bar.start()
            self.progress_message_left.configure(text="")
            self.progress_message_middle.configure(text="Calibration finished!")
            self.progress_message_right.configure(text="")
            self.footer_progress_label.configure(text="Finished")
            self.textbox.insert("0.0", "\nCalibration finished!\n")

def start():
    app = App()
    app.mainloop()

if __name__ == "__main__":
    app = App()
    app.mainloop()


from customtkinter import CTkScrollableFrame
from customtkinter import CTkFrame
from customtkinter import CTkLabel
from customtkinter import CTkButton
from customtkinter import CTkEntry
from customtkinter import CTkOptionMenu
from customtkinter import CTkTextbox
from customtkinter import CTkImage
from .BoundsEditorWindow import BoundsEditorWindow as BEW
from ...util.CTkToolTip import CTkToolTip as ctt
import tkinter as tk
import json
import PIL
from PIL import Image
import os

class BoundsList(CTkFrame):
    def __init__(self, *args,
                 option_manager: None,
                 step_index: 0,
                 **kwargs):
        super().__init__(*args, **kwargs)
        
        self.optParams = []
        self.option_manager = option_manager

        details = self.option_manager.get_service_parameters()
        self.paramMap = {}
        if "parameter" in details:
            for param in details["parameter"]:
                self.paramMap[param["name"]] = param
                if "min" in param:
                    self.optParams.append(param["name"])
                    
            self.optParams.sort()
        

        self.edit_mode = False
        self.tooltip_list = []
        
        self.step_index = step_index 
        self.render()
    
    def clear(self):
        for list in self.tooltip_list:
            for tooltip in list:
                tooltip.destroy()
        self.containerFrame.destroy()
        self.tooltip_list = []
        
    def toggle_edit_mode(self):
        self.clear()
        self.edit_mode = not self.edit_mode
        self.render()
        
    def refresh(self, *args):
        self.clear()
        self.render()
        
    def render(self):
        row = 0
        index = 0
        
        self.tooltip_list = []
        
        self.containerFrame = CTkFrame(self, fg_color="transparent")
        self.containerFrame.grid(row=0, column=0, padx=(5, 5), pady=(5, 5), sticky="nsew")
        self.containerFrame.grid_columnconfigure(0, weight=1, minsize=20)
        self.containerFrame.grid_columnconfigure(5, weight=1, minsize=20)
        self.containerFrame.grid_columnconfigure((1, 2, 3, 4), weight=5)
        
        CTkLabel(self.containerFrame, text="Type:").grid(row=row, column=0, padx=(5, 5), pady=(5, 5), sticky="nsew")
        CTkLabel(self.containerFrame, text="Name:").grid(row=row, column=1, padx=(5, 5), pady=(5, 5), sticky="nsew")
        CTkLabel(self.containerFrame, text="Bounds:").grid(row=row, column=2, columnspan=2, padx=(5, 5), pady=(5, 5), sticky="nsew")
        CTkLabel(self.containerFrame, text="Default:").grid(row=row, column=4, padx=(5, 5), pady=(5, 5), sticky="nsew")
        CTkLabel(self.containerFrame, text="Strategy:").grid(row=row, column=5, padx=(5, 5), pady=(5, 5), sticky="nsew")
        row += 1
        
        bounds = self.option_manager.get_steps()[self.step_index]["param"]
        
        for bound in bounds:
            tt1 = None
            tt2 = None
            tt3 = None
                        
            tt = CTkOptionMenu(self.containerFrame, dynamic_resizing=False, values=['float', 'int', 'list', 'string', 'custom'], variable=bound["type"], command=self.refresh)
            #command = lambda _, index=index, option=tt: (self.update_type(index, option)) 
            #tt.configure(command=command)
            #tt.set(bound["type"].get())
            tt.grid(row=row, column=0, padx=(5, 5), pady=(5, 5), sticky="new")
            
            cc = None
            
            bound_type = bound["type"].get()
            
            if bound_type == "float":
                cc = CTkOptionMenu(self.containerFrame, dynamic_resizing=False, values=self.optParams, variable=bound["name"])
                cc.grid(row=row, column=1, padx=(5, 5), pady=(5, 5), sticky="new")
                
                command = lambda _, index=index: (self.update_values(index), self.update_tooltips(index))  
                cc.configure(command=command)
            else:
                
                cc = CTkEntry(self.containerFrame)
                cc.configure(textvariable=bound["name"])
                cc.grid(row=row, column=1, padx=(5, 5), pady=(5, 5), sticky="new")
                
                        
            if self.edit_mode:
                remove_func = lambda index=index: (self.clear(), self.option_manager.remove_bound(self.step_index, index), self.render())
                bb = CTkButton(self.containerFrame, text="Remove", command=remove_func)
                bb.grid(row=row, column=2, padx=(5, 5), pady=(5, 5), sticky="new")
                ctt(bb, delay=0.1, alpha=0.95, message="Delete this bound...")
            else:
                bounds_min = CTkEntry(self.containerFrame)
                vcmd = (self.register(self.validate_number), '%P', cc, bounds_min)
                bounds_min.grid(row=row, column=2, padx=(5, 5), pady=(5, 5), sticky="new")
                bounds_min.configure(textvariable=bound["bounds"][0], validate='all', validatecommand=vcmd)
                self.validate_number(bounds_min.get(), cc, bounds_min)
                            
                bounds_max = CTkEntry(self.containerFrame)
                vcmd = (self.register(self.validate_number), '%P', cc, bounds_max)
                bounds_max.grid(row=row, column=3, padx=(5, 5), pady=(5, 5), sticky="new")
                bounds_max.configure(textvariable=bound["bounds"][1], validate='all', validatecommand=vcmd)
                self.validate_number(bounds_max.get(), cc, bounds_max)
                tt2 = ctt(bounds_max, delay=0.1, alpha=0.95, message="...")
                
                default_value = CTkEntry(self.containerFrame)
                default_value.grid(row=row, column=4, padx=(5, 5), pady=(5, 5), sticky="new")
                default_value.configure(textvariable=bound["default_value"])
            
                
                if (bound_type == "list"):
                    def button_click_event(bound_index):
                        dialog = BEW(title="Edit List Bound", step_index=self.step_index, bound_index=bound_index, option_manager=self.option_manager)
                        print("Number:", dialog.get_input())

                    open_window = lambda event=None, bound_index=index: (button_click_event(bound_index))
                    expand_image = CTkImage(Image.open(os.path.join("./images", "expand.png")), size=(20, 20))
                    button = CTkButton(self.containerFrame, width=30, text=None, image=expand_image, command=open_window)
                    button.grid(row=row, column=6, padx=(5, 5), pady=(5, 5), sticky="new")
                
                calibration_strat = CTkOptionMenu(self.containerFrame, dynamic_resizing=False, values=['none', 'mean', 'single'], variable=bound["calibration_strategy"])
                calibration_strat.grid(row=row, column=5, padx=(5, 5), pady=(5, 5), sticky="new")
                
                tt1 = ctt(bounds_min, delay=0.1, alpha=0.95, message="...")
                tt2 = ctt(bounds_max, delay=0.1, alpha=0.95, message="...")
                if cc is not None:
                    tt3 = ctt(cc, delay=0.1, alpha=0.95, message="...")
                        
                self.tooltip_list.append([tt3, tt1, tt2])
                        
                self.update_tooltips(index)
            
            row += 1
            index += 1
            
        add_func = lambda: (self.clear(), self.option_manager.add_bound(self.step_index), self.render())
        if len(bounds) > 0:
            if self.edit_mode:
                CTkButton(self.containerFrame, text="Exit", command=self.toggle_edit_mode).grid(row=row, column=0, padx=(5, 5), pady=(5, 5), sticky="new")
            else:
                CTkButton(self.containerFrame, text="Edit", command=self.toggle_edit_mode).grid(row=row, column=0, padx=(5, 5), pady=(5, 5), sticky="new")
            CTkButton(self.containerFrame, text="Add Parameter", command=add_func).grid(row=row, column=1, padx=(5, 5), pady=(5, 5), sticky="new")
        else:
            CTkButton(self.containerFrame, text="Add Parameter", command=add_func).grid(row=row, column=0, columnspan=2, padx=(5, 5), pady=(5, 5), sticky="new")
            
    def update_type(self, index, option):
        value = option.get()
        self.option_manager.get_steps()[self.step_index]["param"][index]["type"].set(value)
        self.refresh()
            
    def update_values(self, index):
        name = self.option_manager.get_steps()[self.step_index]["param"][index]["name"].get()
        if name in self.paramMap:
            obj = self.paramMap[name]
            self.option_manager.get_steps()[self.step_index]["param"][index]["bounds"][0].set(obj["softmin"])
            self.option_manager.get_steps()[self.step_index]["param"][index]["bounds"][1].set(obj["softmax"])
            
    def update_tooltips(self, index):
        try:
            name = self.option_manager.get_steps()[self.step_index]["param"][index]["name"].get()
            bound_type = self.option_manager.get_steps()[self.step_index]["param"][index]["type"].get()
            
            tooltips = self.tooltip_list[index]
            t3 = tooltips[0]
            t1 = tooltips[1]
            t2 = tooltips[2]
            
            if (t1 == None or t2 == None or t3 == None):
                if (t1 is not None):
                    t1.configure(message="")
                if (t2 is not None):
                    t2.configure(message="")
                if (t3 is not None):
                    t3.configure(message="")
                print("update skipped")
                return
            
            if bound_type == "list":
                t1.configure(message="")
                t2.configure(message="")
                t3.configure(message="")
            
            if name in self.paramMap:
                obj = self.paramMap[name]
                description = obj["description"]
                default = str(obj["value"])
                min = str(obj["min"])
                max = str(obj["max"])
                softmin = str(obj["softmin"])
                softmax = str(obj["softmax"])
                unit = str(obj["unit"])
                t1.configure(message=description + "\nMin: " + min + "\nSoft Min: " + softmin + "\nUnit: " + unit)
                t2.configure(message=description + "\nMax: " + max + "\nSoft Max: " + softmax + "\nUnit: " + unit)
                t3.configure(message=description + "\nDefault: " + default + "\nUnit: " + unit)
        except:
            pass
        
    def validate_number(self, P, name, entry):
        
        # Get the root window
        root = self.winfo_toplevel()
        
        # Get the entry widget using its internal Tkinter name
        entry_widget = root.nametowidget(entry)
        name_widget = root.nametowidget(name)
        
        if isinstance(name_widget, CTkTextbox):
            return True
        
        # Call the get method on the entry widget
        bound_name = name_widget.get()
        value = entry_widget.get()
        
        # Print the value of the entry widget
        #print(bound_name)
        
        if P == "" or P == "." or P == "-":
            entry_widget.configure(border_color="red")
            return True
        
        try:
            float(P)
            entry_widget.configure(border_color=["#979DA2", "#565B5E"])
            
            if bound_name in self.paramMap:
                obj = self.paramMap[bound_name]
                if "min" in obj and float(P) < float(obj["min"]):
                    entry_widget.configure(border_color="red")
                elif "max" in obj and float(P) > float(obj["max"]):
                    entry_widget.configure(border_color="red")
                elif "softmin" in obj and float(P) < float(obj["softmin"]):
                    entry_widget.configure(border_color="yellow")
                elif "softmax" in obj and float(P) > float(obj["softmax"]):
                    entry_widget.configure(border_color="yellow")
            
            return True
        except ValueError as e:
            print(e)
            return False
        except Exception as e:
            print(e)
            return False